/*
   DVR, a Digital video Recorder for Linux.
   
   Copyright (C) 2001-2007 Pierre Hebert <pierrox@pierrox.net>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this program; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

 */

#include <string.h>
#include <pthread.h>
#include <gst/gst.h>

#include "dvr.h"

DVR::DVR() : 
  video_filename(NULL),
  device_filename(NULL),
  video_codec(NULL),
  audio_codec(NULL),
  started(false), 
  ended(false)
{
  gst_init(0, NULL);
}

DVR::~DVR() {
  if(started) {
    stop();
  }

  gst_deinit();

  if(device_filename) delete []device_filename;
  if(video_filename) delete []video_filename;
}

void DVR::setImageSize(int w, int h) {
  width=w;
  height=h;
}

void DVR::setFrameRate(int r) {
  framerate=r;
}

void DVR::setVideoFileName(const char *filename) {
  if(video_filename) delete []video_filename;
  video_filename=new char[strlen(filename)+1];
  strcpy(video_filename, filename);
}

void DVR::setDeviceFileName(const char *filename) {
  if(device_filename) delete []device_filename;
  device_filename=new char[strlen(filename)+1];
  strcpy(device_filename, filename);
}

GstElement *DVR::cloneGstElement(GstElement *src, const char *name) {
  if(src==NULL) {
    return NULL;
  }
  
  // duplicate an element.  not sure if this is the right way to do this
  GstElement *dest=gst_element_factory_create(gst_element_get_factory(src), name);

  GParamSpec **property_specs;
  guint num_properties, i;
  GValue value={ 0, { { 0 } } };

  property_specs=g_object_class_list_properties(G_OBJECT_GET_CLASS(src), &num_properties);
  for(i=0; i<num_properties; i++) {
    GParamSpec *param=property_specs[i];
    if(strcmp(param->name, "name") && (param->flags & (G_PARAM_READABLE | G_PARAM_WRITABLE))) {
      g_value_init(&value, param->value_type);
      g_object_get_property(G_OBJECT(src), param->name, &value);
      g_object_set_property(G_OBJECT(dest), param->name, &value);
      g_value_unset(&value);
    }
  }

  return dest;
}

void DVR::setVideoCodec(GstElement *element) {
  video_codec=cloneGstElement(element, "video_codec");
}

void DVR::setAudioCodec(GstElement *element) {
  audio_codec=cloneGstElement(element, "audio_codec");
}

// XXX FIXME TODO object unref
bool DVR::start() {
  context=g_main_context_new();
  loop=g_main_loop_new(NULL, FALSE);

  // TODO conf file : debug mode
  gst_debug_set_active (TRUE);
  gst_debug_set_default_threshold (GST_LEVEL_WARNING);
    
  pipeline=gst_pipeline_new("dvr");
  video_source=gst_element_factory_make("v4l2src", "video_source");
  video_rate=gst_element_factory_make("videorate", "video_rate");
  video_filter=gst_element_factory_make("capsfilter", "video_filter");
  video_convert=gst_element_factory_make("ffmpegcolorspace", "video_convert");
  avimux=gst_element_factory_make("avimux", "avimux");
  file_output=gst_element_factory_make("filesink", "file_output");

  if(!pipeline || !video_source || !video_rate || !video_filter || !video_convert || !video_codec || !avimux || !file_output) {
    return false;
  }
  
  gst_bin_add_many(GST_BIN(pipeline), video_source, video_convert, video_rate, video_filter, video_codec, avimux, file_output, NULL);

  if(audio_codec) {
    audio_source=gst_element_factory_make("alsasrc", "audio_source");
    if(!audio_source) {
      return false;
    }
    gst_bin_add_many(GST_BIN(pipeline), audio_source, audio_codec, NULL);
  }
  

  gst_element_link(video_source, video_rate);

  filter_yuv=gst_caps_new_simple("video/x-raw-yuv",
                                 "width", G_TYPE_INT, width,
                                 "height", G_TYPE_INT, height,
                                 "framerate", GST_TYPE_FRACTION, framerate, 1,
                                 NULL);

  g_object_set(video_filter, "caps", filter_yuv, NULL);

  gst_element_link(video_rate, video_filter);
  gst_element_link(video_filter, video_convert);
  gst_element_link(video_convert, video_codec);
  gst_element_link(video_codec, avimux);
  if(audio_codec) {
    gst_element_link(audio_source, audio_codec);
    gst_element_link(audio_codec, avimux);
  }
  gst_element_link(avimux, file_output);
  
  g_object_set(G_OBJECT(video_source), "device", device_filename, NULL);
  // g_object_set(G_OBJECT(video_source), "autoprobe", FALSE, NULL);
  // g_object_set(G_OBJECT(video_source), "autoprobe-fps", FALSE, NULL);
  g_object_set(G_OBJECT(file_output), "location", video_filename, NULL);

  GstStateChangeReturn res=gst_element_set_state(pipeline, GST_STATE_PLAYING);
  if(res==GST_STATE_CHANGE_FAILURE) {
    // try again with rgb instead of yuv
    filter_rgb=gst_caps_new_simple("video/x-raw-rgb",
                               "width", G_TYPE_INT, width,
                               "height", G_TYPE_INT, height,
                               "framerate", GST_TYPE_FRACTION, framerate, 1,
                               /*"bpp", G_TYPE_INT, 24,
                               "depth", G_TYPE_INT, 24,*/
                               NULL);

    g_object_set(video_filter, "caps", filter_rgb, NULL);

    res=gst_element_set_state(pipeline, GST_STATE_PLAYING);
  }


  if(res!=GST_STATE_CHANGE_FAILURE) {
    pthread_create(&thread_gst, NULL, run, this);
    started=true;
    ended=false;

    return true;
  } else {
    return false;
  }
}

bool DVR::stop() {
  g_main_loop_quit(loop);

  pthread_join(thread_gst, NULL);

  gst_element_set_state(pipeline, GST_STATE_NULL);

  if(pipeline) gst_object_unref(pipeline);
  
  g_main_loop_unref(loop);
  g_main_context_unref(context);
  ended=true;
  started=false;

  return true;
}

bool DVR::isStarted() {
  return started;
}

bool DVR::isEnded() {
  return ended;
}

void *DVR::run(void *data) {
  DVR *me=(DVR*)data;

  g_main_loop_run(me->loop);

  return NULL;
}
